#!/bin/bash
########################################################################
# build_python.sh
#
# Build python and package for UPS.
#
# Adapted from build.sh.template provided by ssibuildshims v2_03_00.
########################################################################

########################################################################
# Non-customize-able boilerplate: skip to "Customize-able functions."
########################################################################

####################################
# Useful variables.
prog=${BASH_SOURCE[0]##*/}

##################
# Local (dumb) ssi_die() function for use until we source ssi_functions.
ssi_die() {
  printf "FATAL ERROR: ${*}\n" 1>&2; exit 1
}
##################

##################
# Basic usage function for use by ensure_buildshims until we have enough
# information to provide something more specific. Do NOT customize!
build_script_usage() {
    cat <<EOF
USAGE: ${prog} <ssi-msg-config-options> [--] <product-dir> <qualifier-args> <maketar>
       ${prog} -h|--help [<product-dir>]

  Build ${package} and package for UPS.

OPTIONS

  -h|--help

    Help. Supply <product-dir> for specific, detailed information on
    invoking ${prog}.

  <ssi-msg-config-options>

    Supply <product-dir> for more information.

ARGUMENTS

  <product-dir>

    The top-level UPS directory in which the source code distribution
    was unpacked. When specified with -h|--help, this enables us to
    provide more specific help about invoking ${prog}.

  <qualifier-args>
  <maketar>

    Supply <product-dir> for more information.

EOF
}
##################

##################
# Make sure we have UPS and ssibuildshims set up.
ensure_ssibuildshims() {
  if ! (( $# )) || [ "${@: -1}" = "-h" ] || [ "${@: -1}" = "--help" ]; then
    build_script_usage
    exit 1
  fi
  # Find <product-dir> from our arguments:
  local arg
  for arg in "${@}"; do
    [[ "$arg" ==  -* ]] && continue;
    product_dir="${arg}"; break
  done
  if [ -z "${product_dir}" ]; then
    # Bad usage.
    ssi_die "Bad <product-dir> ${product_dir}\n$(build_script_usage)"
    exit 2
  elif ! { [ -d "${product_dir}/.upsfiles" ] && [ -w "${product_dir}" ]; }; then
    # Something wrong with <product-dir>.
    ssi_die "<product-dir> not a valid, writable unified UPS directory: " \
      "${product_dir}\n$(build_script_usage)"
  fi
  # Promising...
  local abs_product_dir="$(cd "${product_dir}" 2>/dev/null 2>&1 && pwd -P)"
  (( $? )) && ssi_die "unable to cd to <product-dir>: ${product_dir}\n" \
    "$(build_script_usage)"
  # Find our own bootstraps and yank 'em.
  local ssibuildshims_topdir="${product_dir}/ssibuildshims/${ssibuildshims_version}"
  [ -f "${ssibuildshims_topdir}/bin/ssi_functions" ] && \
    source "${ssibuildshims_topdir}/bin/ssi_functions" >/dev/null 2>&1 && \
    ensure_ups "${product_dir}" && \
    setup ssibuildshims ${ssibuildshims_version} -z "${product_dir}" || \
    ssi_die "unable to find ssibuildshims ${ssibuildshims_version} in
${product_dir} and set it up with UPS."
  # Absolute is safer.
  absolute_path -- "${product_dir}" product_dir
}
##################

########################################################################
# Customize-able functions.
########################################################################

##################
# Ascertain whether we built and installed correctly. Add more /
# different calls to verify_in_filesystem as appropriate. Arguments
# (e.g. -q) are passed through to verify_in_filesystem.
is_built() { # *C*
  ${trace} verify_in_filesystem -x "${@}" "${pkgdir}/bin/python"{,3,2}
}
##################

##################
# If a distribution archive is wanted and the installation is successful
# according to is_built(), then make it. Any arguments (e.g. -q) are
# passed through both to is_built() and make_basic_distribution().
#
# If make_basic_distribution() finds that the product is declared
# correctly and attempts to make the distribution, then it will cause us
# to exit with the status code of that operation. If not, it returns an
# error code.
#
# See make_basic_distribution -h for useful options to add to
# mbd_opts(). *C*
maybe_make_distribution() {
  if [ "${maketar}" = "tar" ]; then
    # Customize mbd_opts. *C*
    local mbd_opts=(-x scripts) # Scripts are all build-time utilities.
    ${trace} is_built "$@" && \
      ${trace} make_basic_distribution "${@}" "${mbd_opts[@]}"
  fi
}
##################

########################################################################
# Main.
########################################################################

####################################
# Useful variables. *C*
thisdir=$(cd "$(dirname "${BASH_SOURCE[0]}")" && pwd -P)
trace=maybe_xtrace # Could set to xtrace instead for unconditional.
(( ssi_trace = 1 )) # Trace certain commands by default.
####################################

####################################
# Package name and version information. *C*
. ${thisdir}/version_info.sh || \
  ssi_die "unable to load version info from ${thisdir}/version_info"
####################################

##################
# Other useful variables.
pip_version=22.1
setuptools_version=62.2.0
wheel_version=0.37.1
##################

####################################
# Handle options and arguments.

# Set up ssibuildshims and UPS / product_dir as early as possible.
ensure_ssibuildshims "${@}"

##################
# Now parse options and arguments, setting basequal, extraqual, maketar,
# etc.
#
# Use verify_quals() to parse command line options, verify that the user
# has provided qualifiers supported by this build script, and set
# expected variables such as basequal, extraqual, cqual, etc.
#
# Customize the extra usage text as appropriate. Note that DEBUG_PATCHES
# and VERBOSE_PATCHES are honored by prepare_source() and
# patch_source(). DEBUG_(CONFIG|BUILD|TEST|INSTALL) should be accounted
# for (or not) by the relevant command in this script.
#
# See verify_quals -h for options and examples.
verify_quals \
  --parse --usage-file=- -- "$@" <<EOF || exit # *C*
ENVIRONMENT

  DEBUG_PATCHES
  VERBOSE_PATCHES

    Command tracing and increased verbosity for source unpacking and patching
    operations. DEBUG_PATCHES will cause exit after patches have been applied.

  PYTHON_WANT_TESTS

    Activate tests. Since Python's tests are extensive, time-consuming,
    and problematic in several environments, we use a package-specific
    environment variable rather than the conventional, "WANT_TESTS."
EOF
##################

##################
# Define globals: pkgdir, sourcedir, blddir, fullqual, etc. See
# define_basics -h for steering options.
${trace} define_basics NO_QUAL NO_DEBUG_SOURCE TRACK_PROGRESS || exit # *C*

# Our own dependent variables.
scriptsdir="${pkgtop}/scripts"
python_pkgs_file="${scriptsdir}/requirements.txt"
##################

####################################
# Declare ourselves privately, and then set ourselves up to allow setup
# of dependencies. *C*
${trace} fake_declare_product -x -- || exit

# Can't have PYTHONHOME, etc., lying around while we build.
unsetup -j "${package}"
####################################

####################################
# Build-only dependencies. *C*
#
# e.g.
#   setup_cmake --latest || exit
# OR
#   setup_ninja --optional && (( want_ninja = 1 ))
####################################

# Print active products.
echo
ups active
echo

# Print environment.
printenv | sort

# Possible short-circuit if we're already built and declared.
${trace} maybe_make_distribution --silent

ssi_info "building ${package} for ${OS}-${plat}-${qualdir} (flavor ${pkgflvr})"

# Ensure directories: pkgdir, sourcedir, blddir.
${trace} init_pkg_dirs || exit

####################################
# Prepare source for building. *C*

${trace} cd "${sourcedir}" \
  || ssi_die "unable to cd to source directory ${sourcedir}"

##################
# Source tarballs and patches. See prepare_source -h / patch_source -h
# for options.

declare -a patches=("${pkgtarname}.patch")

[ ${OS1} = "Linux" ] && [ "$(lsb_release -i -s)" = "SUSE LINUX" ] && \
  patches+=("${patchdir}/SUSE.Modules.panel.patch")

${trace} prepare_source "${ssi_args[@]}" -P "${patchdir}" \
  "${tardir}/${pkgtarfile}" \
  "${patches[@]}" \
  || exit

# Bail early if DEBUG_PATCHES is set.
[ -n "${DEBUG_PATCHES}" ] && exit 1
####################################

####################################
# Prepare to build. *C*
#
# Set configure / compile flags, environment variables, etc.

configure_opts=(--prefix="${pkgdir}"
  --enable-ipv6
  --without-ensurepip
  --enable-shared
  --enable-loadable-sqlite-extensions
  --enable-big-digits
  --with-system-expat
  --with-system-ffi
  --with-readline=editline
  --without-static-libpython
  )

if [ "${OS1}" = Darwin ]; then
  tpossl=/usr/local/opt/openssl
  if [ -d "${tpossl}" ]; then
    configure_opts+=(--with-openssl="${tpossl}" --with-ssl-default-suites=openssl)
  else
    ssi_warn "unable to find third party OpenSSL in ${tpossl}: building without OpenSSL"
  fi

  configure_opts+=(--without-gcc --enable-optimizations --with-lto)
  export CC=clang
  export CXX=clang++
elif ! ver_greater 6.1 "$(gcc -dumpversion)"; then
  configure_opts+=(--enable-optimizations --with-lto)
fi

##################
# Configure parallelism. *C*
ncores=$(ncores)
##################
####################################

####################################
# Configure, build, and install.
#
# Use ${blddir}/${pkgtarname} for in-source builds, ${blddir} otherwise.

${trace} cd "${blddir}" || \
  ssi_die "unable to cd to build directory ${blddir}"

##################
# Configure. *C*

${trace} "${sourcedir}/${pkgtarname}/configure" "${configure_opts[@]}" || \
  ssi_die "configure failed"
##################

##################
# Build. *C*

${trace} make -j ${ncores} || ssi_die "build failed"
##################

##################
# Test. *C*

(( PYTHON_WANT_TESTS )) && \
  { ${trace} make -j ${ncores} buildbottest || ssi_die "tests failed"; }
##################

##################
# Install. *C*

${trace} make -j ${ncores} install || ssi_die "install failed"
##################

ssi_info "finished building ${package} ${pkgver}"

####################################
# Pre-declaration processing. *C*
#
# Move things, install from contrib, clean build products from in-source
# builds, etc.
####################################

ssi_info -i "post-processing for relocatability (1/2)"

sysconfig_py=($(find "${pkgdir}" -name '_sysconfigdata*.py'))

absolute_path -- "${SQLITE_FQ_DIR}" sqlfq

[ -n "${sysconfig_py}" ] || \
  ssi_die "could not find sysconfig data file to fix"

${trace} perlfilt -wpi'' \
  -e 'm&^build_time_vars& and print <<'"'"'EOF'"'"';
from os import environ as env
this_python_root=env["PYTHON_ROOT"]
this_python_dir=env["PYTHON_DIR"]
this_sqlite_fq_dir=env["SQLITE_FQ_DIR"]
EOF
' \
  -e "/'CONFIG_ARGS':/ and do { s&\\Q${pkgdir}\\E&\"+this_python_root+\"&g;
s&\\Q${sqlfq}\\E&\"+this_sqlite_fq_dir+\"&g; };" \
  -e "/'PKG_CONFIG_PATH=/ and do { s&\\Q${pkgdir}\\E&\"+this_python_root+\"&g;
s&\\Q${sqlfq}\\E&\"+this_sqlite_fq_dir+\"&g; };" \
  -e "/'(?:DY)?LD_LIBRARY_PATH=/ and do { s&\\Q${pkgdir}\\E&'+this_python_root+'&g;
s&\\Q${sqlfq}\\E&'+this_sqlite_fq_dir+'&g; };" \
  -e "s&'\\Q${pkgdir}\\E'&this_python_root&g;" \
  -e "s&'\\Q${pkgdir}\\E&this_python_root+'&g;" \
  -e "s&\\Q${pkgdir}\\E&'+this_python_root'&g;" \
  -e "s&-L\\Q${sqlfq}\\E&-L'+this_sqlite_fq_dir+'&g;" \
  -e "s&\\Q${sqlfq}\\E&'+this_sqlite_fq_dir+'&g;" \
  -e "s&^(\s+)this_&\${1}+this_&g;" \
  "${sysconfig_py}" && \
  ! ${trace} grep -Fe "${pkgdir}" -e "${sqlfq}" -e "${SQLITE_FQ_DIR}" \
  "${sysconfig_py}" >/dev/null 2>&1 || \
  ssi_die "unable to fix ${sysconfig_py}"
{ set +x; } 2>/dev/null
absolute_path -- "${SQLITE_LIB}" sqllib
absolute_path -- "${SQLITE_INC}" sqlinc

makefiles=($(find "${pkgdir}/lib" -name Makefile))
${trace} perlfilt -wpi'' \
  -e 's&\Q'"${pkgdir}/include/${PYTHON_LIBDIR}"'\E&\$(PYTHON_INCLUDE)&g;' \
  -e 's&\Q'"${pkdir}/lib"'\E&\$(PYTHON_LIB)&g;' \
  -e 's&\Q'"${pkgdir}"'\E&\$(PYTHONHOME)&g;' \
  -e 's&(LD_LIBRARY_PATH=|:)\Q'"${blddir}"'\E:?&${1}&;' \
  -e 's&\Q'"${pkgtop}"'\E&\$(PYTHON_DIR)&g;' \
  -e 's&\Q'"${sqlinc}"'\E&\$(SQLITE_INC)&g;' \
  -e 's&\Q'"${sqllib}"'\E&\$(SQLITE_LIB)&g;' \
  -e 's&\Q'"${sqlfq}"'\E&\$(SQLITE_FQ_DIR)&g;' \
  "${makefiles[@]}" && \
  ! ${trace} grep -Fe "${pkgdir}" -e "${sqlfq}" -e "${SQLITE_FQ_DIR}" \
  "${makefiles[@]}" >/dev/null 2>&1 || \
  ssi_die "unable to fix ${makefiles[*]}"

####################################
# N.B. It is *not* necessary to fix python*-config: the original
# hard-wired prefix is kept for reference, but it gives the correct
# information based on where it is installed.
####################################

${trace} fix_pc ${pkgdir}/lib/pkgconfig/*.pc || \
  ${trace} smoke_test_pc ${pkgdir}/lib/pkgconfig/*.pc || \
  ssi_die "unable to fix pkg-config files."

##################
# Declare and set up the installed package. *C*
#
# Add -c for current declaration if required.
${trace} declare_product -x -- || exit
##################

####################################
# Post-declaration processing. *C*
#
# Operations requiring the *final* declaration and setup (as opposed to
# the pre-build fake declaration and setup). This should be rare.
#
# Python is rare, thankfully.
####################################

${trace} python_vars "${pkgdir}/bin/python3" \
  || ssi_build_failed "unable to ascertain python vars for fresh-built python"

ssi_info "checking expected Python Standard Library modules"
${trace} ${python} "${scriptsdir}/test-standard-modules.py" || \
  ssi_die "Python Standard Library checks failed"

ssi_info "installing recent pip and common Python packages"
get_pip_url="https://bootstrap.pypa.io/get-pip.py"
${trace} curl "${get_pip_url}" --silent --fail | \
  ${python} - --no-cache-dir --no-wheel --no-setuptools \
  --prefix "${pkgdir}" || \
  ssi_die "unable to obtain pip using ${get_pip_url}"

latest_pip_version=$(pip -V | cut -d' ' -f2)

if ver_greater ${latest_pip_version} ${pip_version}; then
  ssi_info "using pip ${latest_pip_version} to install known version " \
    "${pip_version} of pip"
  pip install -I pip==${pip_version} || \
    ssi_die "failed to install required pip ${pip_version}"
fi

ssi_info "installing setuptools ${setuptools_version}"
${trace} pip install setuptools==${setuptools_version} || \
  ssi_die "Failed to install setuptools ${setuptools_version}"

ssi_info "installing wheel ${wheel_version}"
${trace} pip install wheel==${wheel_version} || \
  ssi_die "Failed to install wheel ${wheel_version}"

ssi_info "installing Python packages specified in ${python_pkgs_file}"
${trace} pip install --no-cache-dir -r "${python_pkgs_file}" \
  --prefix "${pkgdir}" || \
  ssi_die "Failed executing 'pip install -r \"${python_pkgs_file}\"'"

ssi_info "post-processing for relocatability (2/2)"
failed_fix=()
for file in "${pkgdir}/bin"/* "${pkgdir}/lib/"*/*/python-config.py; do
  file_basename="${file##*/}"
  if [[ "${file_basename}" =~ ^(.*)(-)?${pythonMdotm}(-config)?$ ]]; then
    for link in \
      "${file%/*}/${BASH_REMATCH[1]}"{${BASH_REMATCH[2]}${pythonMdotm%.*},}"${BASH_REMATCH[3]}"; do
      [ -f "${link}" ] || \
        ${trace} ln -sv "${file_basename}" "${link}" || \
        ssi_error -i "unable to make link from ${file_basename} to ${link##*/}\n" \
        "in ${link%/*}" || \
        failed_fix+=("${link}")
    done
  fi
  [ -L "${file}" ] || ! grep -IEe '^\#!' "${file}" >/dev/null 2>&1 || \
    ${trace} sed -Ei'' -e \
    '1s&^#!.*(python([0-9]+(\.[0-9]+)?)?)($|[[:space:]])&#!/usr/bin/env python\2\4&' \
    -e 's&/.*/(python([0-9]+(\.[0-9]+)?)?)($|[[:space:]])&python\2\4&' \
    "${file}"
done
(( ${#failed_fix[@]} == 0 )) || \
  ssi_die "post-processing 2/2 failed: unable to fix ${#failed_fix[@]} files:" \
  "${failed_fix[@]/#/\n  }"
####################################

#################
# Verify installation. *C*
#
# To customize verification, see is_built(), since it is also used by
# maybe_make_distribution(). See the documentation for those functions
# (above) for customization opportunities.

# Add a message to ssi_build_failed if appropriate.
${trace} is_built && \
  ${trace} ssi_build_completed && \
  ssi_info "${package} is installed at ${pkgdir}" && \
  ${trace} maybe_make_distribution || \
  ssi_build_failed # *C*
########################################################################
# Nothing beyond this point.
########################################################################
